import mimetypes
import logging
import os.path
import requests

from tqdm import tqdm


def read_in_chunks(file_path, blocksize=1024, chunks=-1):
    """ Splitting the file into chunks. """
    with open(file_path, 'rb') as file_object:
        size = os.path.getsize(file_path)
        pbar = tqdm(total=100)
        current = 0
        while chunks:
            data = file_object.read(blocksize)
            if not data:
                break
            yield data
            chunks -= 1
            step = round(blocksize / size * 100, 1)
            current = min(100, current + step)
            pbar.update(min(100 - current, step))
        pbar.update(100 - current)
        pbar.close()


def upload_to_signed_url(file_path, signed_url, querier, orm_class):
    res_upload = requests.put(
        signed_url.get('signed_url'),
        data=read_in_chunks(file_path),
        headers={'Content-Type': 'application/octet-stream'}
    )
    res = None
    if res_upload.status_code == 200:
        data_hash = signed_url.get('data_hash')

        res = querier.basic_put(
            orm_class,
            uid=data_hash,
            payload=signed_url
        )

        if orm_class(res):
            logging.info("Successfully uploaded: %s",
                         signed_url.get('title', ''))
        else:
            logging.info("Error uploading: %s",
                         signed_url.get('title', ''))

    else:
        logging.info("Error generating signed url for: %s",
                     signed_url.get('title', ''))

    return orm_class(res)


def upload_to_signed_url_list(file_paths, signed_urls, querier, orm_class):
    assert len(file_paths) == len(signed_urls),\
        'Error getting the correct number of signed urls'
    data_uid_list = []
    for i in range(len(file_paths)):
        file_path = file_paths[i]
        file_name = os.path.basename(file_path)
        mime_type = mimetypes.guess_type(file_path)[0]
        url = signed_urls[i]
        assert url.get('title', '') == file_name, 'Ordering issue'
        res_upload = requests.put(
            url.get('signed_url'),
            data=read_in_chunks(file_path),
            headers={'Content-Type': mime_type}
        )
        if res_upload.status_code == 200:
            data_hash = url.get('data_hash')
            res = querier.basic_put(
                orm_class,
                uid=data_hash,
                payload=url
            )
            if res:
                logging.info("Successfully uploaded: %s",
                             url.get('title', ''))
                data_uid_list.append(url.get('data_hash'))
            else:
                logging.info("Error uploading: %s",
                             url.get('title', ''))
                raise Exception('Could not save information into database')
        else:
            raise Exception('Bad request')

    return data_uid_list





from dataclasses import dataclass

from cord.project_ontology.object_type import ObjectShape


@dataclass
class OntologyObject:
    id: str
    color: str
    name: str
    shape: ObjectShape
    feature_node_hash: str




#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import dataclasses
import json
from collections import OrderedDict
from enum import IntEnum, Enum
from typing import List, Dict

from cord.orm import base_orm
from cord.orm.formatter import Formatter


class Dataset(base_orm.BaseORM):
    """
    A dataset is a collection of data rows.

    ORM:

    title,
    description,
    dataset_type (Cord storage vs. AWS/GCP/Azure),
    data_rows: [
        {
            data_hash (uid),
            data_title,
            data_type,
        }
    ]

    """

    DB_FIELDS = OrderedDict([
        ("title", str),
        ("description", str),
        ("dataset_type", str),
        ("data_rows", (list, str))
    ])

    NON_UPDATABLE_FIELDS = {
        "dataset_type",
    }


@dataclasses.dataclass(frozen=True)
class DatasetAPIKey(Formatter):
    dataset_hash: str
    api_key: str
    title: str
    key_hash: str
    scopes: List[DatasetScope]

    @classmethod
    def from_dict(cls, json_dict: Dict):
        if isinstance(json_dict['scopes'], str):
            json_dict['scopes'] = json.loads(json_dict['scopes'])
        scopes = [DatasetScope(scope) for scope in json_dict['scopes']]
        return DatasetAPIKey(json_dict['resource_hash'], json_dict['api_key'], json_dict['title'],
                             json_dict['key_hash'], scopes)


class DatasetType(IntEnum):
    CORD_STORAGE = 0,
    AWS = 1,
    GCP = 2,
    AZURE = 3


class DatasetScope(Enum):
    READ = 'dataset.read'
    WRITE = 'dataset.write'


class DatasetData(base_orm.BaseORM):
    """
    Video base ORM.
    """

    DB_FIELDS = OrderedDict([
        ("data_hash", str),
        ("video", dict),
        ("images", list),
    ])


class SignedVideoURL(base_orm.BaseORM):
    """ A signed URL object with supporting information. """
    DB_FIELDS = OrderedDict([
        ("signed_url", str),
        ("data_hash", str),
        ("title", str),
        ("file_link", str)
    ])


class SignedImageURL(base_orm.BaseORM):
    """ A signed URL object with supporting information. """
    DB_FIELDS = OrderedDict([
        ("signed_url", str),
        ("data_hash", str),
        ("title", str),
        ("file_link", str)
    ])


class SignedImagesURL(base_orm.BaseListORM):
    """ A signed URL object with supporting information. """
    BASE_ORM_TYPE = SignedImageURL


class Video(base_orm.BaseORM):
    """ A video object with supporting information. """
    DB_FIELDS = OrderedDict([
        ("data_hash", str),
        ("title", str),
        ("file_link", str),
    ])

    NON_UPDATABLE_FIELDS = {
        "data_hash",
    }


class ImageGroup(base_orm.BaseORM):
    """ An image group object with supporting information. """
    DB_FIELDS = OrderedDict([
        ("data_hash", str),
        ("title", str),
        ("file_link", str),
    ])

    NON_UPDATABLE_FIELDS = {
        "data_hash",
    }


class Image(base_orm.BaseORM):
    """ An image object with supporting information. """
    DB_FIELDS = OrderedDict([
        ("data_hash", str),
        ("title", str),
        ("file_link", str),
    ])

    NON_UPDATABLE_FIELDS = {
        "data_hash",
    }


@dataclasses.dataclass(frozen=True)
class ReEncodeVideoTaskResult:
    data_hash: str
    signed_url: str
    bucket_path: str


@dataclasses.dataclass(frozen=True)
class ReEncodeVideoTask:
    """ A re encode video object with supporting information. """
    status: str
    result: List[ReEncodeVideoTaskResult] = None



#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cord.constants.model import FAST_AI, YOLOV5, FASTER_RCNN, MASK_RCNN
from cord.orm.model import ModelTrainingWeights

fast_ai = ModelTrainingWeights({
    "model": FAST_AI,
    "training_config_link": "",
    "training_weights_link": ""
})

yolov5x = ModelTrainingWeights({
    "model": YOLOV5,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Yolo/yolov5x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Yolo/yolov5x.pt"
})

yolov5s = ModelTrainingWeights({
    "model": YOLOV5,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Yolo/yolov5s.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Yolo/yolov5s.pt"
})

faster_rcnn_R_50_C4_1x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_R_50_C4_1x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_721ade.pkl"
})

faster_rcnn_R_50_DC5_1x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_R_50_DC5_1x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_51d356.pkl"
})

faster_rcnn_R_50_FPN_1x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_R_50_FPN_1x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_b275ba.pkl"
})

faster_rcnn_R_50_C4_3x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_R_50_C4_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_f97cb7.pkl"
})

faster_rcnn_R_50_DC5_3x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_R_50_DC5_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_68d202.pkl"
})

faster_rcnn_R_50_FPN_3x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_R_50_FPN_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_280758.pkl"
})

faster_rcnn_R_101_C4_3x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_R_101_C4_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_298dad.pkl"
})

faster_rcnn_R_101_DC5_3x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_R_101_DC5_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_3e0943.pkl"
})

faster_rcnn_R_101_FPN_3x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_R_101_FPN_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_f6e8b1.pkl"
})

faster_rcnn_X_101_32x8d_FPN_3x = ModelTrainingWeights({
    "model": FASTER_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/faster_rcnn_X_101_32x8d_FPN_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_68b088.pkl"
})

mask_rcnn_X_101_32x8d_FPN_3x = ModelTrainingWeights({
    "model": MASK_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/mask_rcnn_X_101_32x8d_FPN_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_2d9806.pkl"
})

mask_rcnn_R_50_C4_1x = ModelTrainingWeights({
    "model": MASK_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/mask_rcnn_R_50_C4_1x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_9243eb.pkl"
})

mask_rcnn_R_50_C4_3x = ModelTrainingWeights({
    "model": MASK_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/mask_rcnn_R_50_C4_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_4ce675.pkl"
})

mask_rcnn_R_101_FPN_3x = ModelTrainingWeights({
    "model": MASK_RCNN,
    "training_config_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/config/mask_rcnn_R_101_FPN_3x.yaml",
    "training_weights_link": "https://cord-models.s3.eu-west-2.amazonaws.com/Detectron/weights/model_final_a3ec72.pkl"
})



#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pprint
from enum import Enum


def pretty_print(data):
    return pprint.pformat(data, indent=4, width=10)


class APIKeyScopes(Enum):
    LABEL_READ = "label.read"
    LABEL_WRITE = "label.write"
    MODEL_INFERENCE = "model.inference"
    MODEL_TRAIN = "model.train"
    LABEL_LOGS_READ = "label_logs.read"
    ALGO_LIBRARY = "algo.library"




#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import hashlib
import logging
import os
from abc import ABC, abstractmethod
from typing import Dict, Optional

import cryptography
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey
from cryptography.hazmat.primitives.serialization import load_ssh_private_key, Encoding, PublicFormat

import cord.exceptions

CORD_ENDPOINT = 'https://api.cord.tech/public'
CORD_USER_ENDPOINT = 'https://api.cord.tech/public/user'
WEBSOCKET_ENDPOINT = 'wss://message-api.cord.tech/websocket'

_CORD_PROJECT_ID = 'CORD_PROJECT_ID'
_CORD_DATASET_ID = 'CORD_DATASET_ID'
_CORD_API_KEY = 'CORD_API_KEY'

READ_TIMEOUT = 180  # In seconds
WRITE_TIMEOUT = 180  # In seconds
CONNECT_TIMEOUT = 180  # In seconds

log = logging.getLogger(__name__)


class BaseConfig(ABC):
    def __init__(self, endpoint: str):
        self.read_timeout: int = READ_TIMEOUT
        self.write_timeout: int = WRITE_TIMEOUT
        self.connect_timeout: int = CONNECT_TIMEOUT

        self.endpoint: str = endpoint

    @abstractmethod
    def define_headers(self, data: str) -> Dict:
        pass


class Config(BaseConfig):
    """
    Config defining endpoint, project id, API key, and timeouts.
    """

    def define_headers(self, data) -> Dict:
        return self._headers

    def __init__(self,
                 resource_id: str = None,
                 api_key: str = None,
                 endpoint: str = CORD_ENDPOINT,
                 websocket_endpoint: str = WEBSOCKET_ENDPOINT
                 ):

        if resource_id is None:
            resource_id = get_env_resource_id()

        if api_key is None:
            api_key = get_env_api_key()

        self.resource_id = resource_id
        self.api_key = api_key
        self.endpoint = endpoint
        self.websocket_endpoint = websocket_endpoint
        self._headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'ResourceID': resource_id,
            'Authorization': self.api_key
        }

        super().__init__(endpoint)
        log.info("Initialising Cord client with endpoint: %s and resource_id: %s",
                 endpoint, resource_id)


def get_env_resource_id() -> str:
    if (_CORD_PROJECT_ID in os.environ) and (_CORD_DATASET_ID in os.environ):
        raise cord.exceptions.InitialisationError(
            message=(
                "Found both Project ID and Dataset ID in os.environ. "
                "Please initialise CordClient by passing resource_id."
            )
        )

    elif _CORD_PROJECT_ID in os.environ:
        resource_id = os.environ[_CORD_PROJECT_ID]

    elif _CORD_DATASET_ID in os.environ:
        resource_id = os.environ[_CORD_DATASET_ID]

    else:
        raise cord.exceptions.AuthenticationError(
            message="Project ID or dataset ID not provided"
        )

    return resource_id


def get_env_api_key() -> str:
    if _CORD_API_KEY not in os.environ:
        raise cord.exceptions.AuthenticationError(
            message="API key not provided"
        )

    return os.environ[_CORD_API_KEY]


class CordConfig(Config):
    def __init__(self, resource_id=None, api_key=None):
        super(CordConfig, self).__init__(resource_id, api_key)


class UserConfig(BaseConfig):
    def __init__(self, private_key: Ed25519PrivateKey, endpoint: str = CORD_USER_ENDPOINT):
        self.private_key: Ed25519PrivateKey = private_key
        self.public_key: Ed25519PublicKey = private_key.public_key()
        self._public_key_hex: str = self.public_key.public_bytes(Encoding.Raw, PublicFormat.Raw).hex()

        super().__init__(endpoint)

    def define_headers(self, data: str) -> Dict:
        hash_builder = hashlib.sha256()
        hash_builder.update(data.encode())
        contents_hash = hash_builder.digest()

        signature = self.private_key.sign(contents_hash)

        return {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': f'{self._public_key_hex}:{signature.hex()}'
        }

    @staticmethod
    def from_ssh_private_key(ssh_private_key: str, password: Optional[str], **kwargs):
        key_bytes = ssh_private_key.encode()
        password_bytes = password and password.encode()
        private_key = cryptography.hazmat.primitives.serialization.load_ssh_private_key(key_bytes, password_bytes)

        if isinstance(private_key, Ed25519PrivateKey):
            return UserConfig(private_key, **kwargs)
        else:
            raise ValueError(f'Provided key [{ssh_private_key}] is not an Ed25519 private key')



#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cord.exceptions import *

# Error messages
AUTHENTICATION_ERROR = ['AUTHENTICATION_ERROR']
AUTHORISATION_ERROR = ['AUTHORISATION_ERROR']
METHOD_NOT_ALLOWED_ERROR = ['METHOD_NOT_ALLOWED_ERROR']
UNKNOWN_ERROR = ['UNKNOWN_ERROR']
OPERATION_NOT_ALLOWED_ERROR = ['OPERATION_NOT_ALLOWED']
ANSWER_DICTIONARY_ERROR = ['ANSWER_DICTIONARY_ERROR']
CORRUPTED_LABEL_ERROR = ['CORRUPTED_LABEL_ERROR']
FILE_TYPE_NOT_SUPPORTED_ERROR = ['FILE_TYPE_NOT_SUPPORTED_ERROR']
FEATURE_DOES_NOT_EXIST_ERROR = ['FEATURE_DOES_NOT_EXIST_ERROR']
MODEL_WEIGHTS_INCONSISTENT_ERROR = ['MODEL_WEIGHTS_INCONSISTENT_ERROR']
MODEL_FEATURES_INCONSISTENT_ERROR = ['MODEL_FEATURES_INCONSISTENT_ERROR']
UPLOAD_OPERATION_NOT_SUPPORTED_ERROR = ['UPLOAD_OPERATION_NOT_SUPPORTED_ERROR']
MUST_SET_DETECTION_RANGE_ERROR = ['MUST_SET_DETECTION_RANGE_ERROR']
DETECTION_RANGE_INVALID_ERROR = ['DETECTION_RANGE_INVALID_ERROR']
RESOURCE_EXISTS_ERROR = ['RESOURCE_EXISTS_ERROR']


def check_error_response(response, payload=None):
    """
    Checks server response.
    Called if HTTP response status code is an error response.
    """
    if response == AUTHENTICATION_ERROR:
        raise AuthenticationError("Invalid API key.")

    if response == AUTHORISATION_ERROR:
        raise AuthorisationError("You are not authorised to access this asset.")

    if response == METHOD_NOT_ALLOWED_ERROR:
        raise MethodNotAllowedError("HTTP method is not allowed.")

    if response == UNKNOWN_ERROR:
        raise UnknownException("An unknown error occurred.")

    if response == OPERATION_NOT_ALLOWED_ERROR:
        raise OperationNotAllowed("The read/write operation is not allowed by the API key.")

    if response == ANSWER_DICTIONARY_ERROR:
        raise AnswerDictionaryError("An object or classification is missing in the answer dictionaries.")

    if response == CORRUPTED_LABEL_ERROR:
        raise CorruptedLabelError("The label blurb is corrupted. This could be due to the number of "
                                  "frame labels exceeding the number of frames in the labelled video.")

    if response == FILE_TYPE_NOT_SUPPORTED_ERROR:
        raise FileTypeNotSupportedError("Supported file types are: image/jpeg, image/png, video/webm, video/mp4.")

    if response == FEATURE_DOES_NOT_EXIST_ERROR:
        raise FeatureDoesNotExistError("The passed feature does not exist in the project ontology.")

    if response == MODEL_WEIGHTS_INCONSISTENT_ERROR:
        raise ModelWeightsInconsistentError("The passed model weights are incompatible with the selected model.")

    if response == MODEL_FEATURES_INCONSISTENT_ERROR:
        raise ModelFeaturesInconsistentError("The passed features are incompatible with the selected model.")

    if response == UPLOAD_OPERATION_NOT_SUPPORTED_ERROR:
        raise UploadOperationNotSupportedError("Uploading a file to an external (e.g. S3/GCP/Azure) dataset is not "
                                               "permitted.")

    if response == DETECTION_RANGE_INVALID_ERROR:
        raise DetectionRangeInvalidError("The detection range is invalid (e.g. less than 0, or"
                                         " higher than num frames in the video)")

    if response == RESOURCE_EXISTS_ERROR:
        raise ResourceExistsError(
            "Trying to create a resource that already exists. "
            "Payload for this failure is: " + str(payload)
        )

    pass



#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

""" ``cord.client`` provides a simple Python client that allows you
to query project resources through the Cord API.

Here is a simple example for instantiating the client for a project
and obtaining project info:

.. test_blurb2.py code::

    from cord.client import CordClient

    client = CordClient.initialize('YourProjectID', 'YourAPIKey')
    client.get_project()

    Returns:
        Project: A project record instance. See Project ORM for details.

"""

from __future__ import annotations

import base64
import json
import logging
import os.path
import sys
import typing
import uuid
from pathlib import Path
from typing import List, Tuple, Union, Optional

import cord.exceptions
from cord.configs import CordConfig, Config
from cord.constants.model import *
from cord.constants.string_constants import *
from cord.http.querier import Querier
from cord.http.utils import upload_to_signed_url, upload_to_signed_url_list
from cord.orm.api_key import ApiKeyMeta
from cord.orm.cloud_integration import CloudIntegration
from cord.orm.dataset import (
    Dataset, Image, ImageGroup, SignedImagesURL, SignedVideoURL, Video, DatasetData, ReEncodeVideoTask
)
from cord.orm.label_log import LabelLog
from cord.orm.label_row import LabelRow
from cord.orm.labeling_algorithm import (
    LabelingAlgorithm, ObjectInterpolationParams, BoundingBoxFittingParams
)
from cord.orm.model import Model, ModelRow, ModelInferenceParams, ModelTrainingParams, ModelOperations
from cord.orm.project import Project, ProjectCopy, ProjectDataset, ProjectUsers

from cord.project_ontology.classification_type import ClassificationType
from cord.project_ontology.object_type import ObjectShape
from cord.project_ontology.ontology import Ontology
from cord.utilities.project_user import ProjectUserRole, ProjectUser

# Logging configuration
logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='[%(asctime)s] [%(levelname)s] [%(filename)s:%(lineno)s] [%(funcName)s()] %(message)s',
                    datefmt='%m/%d/%Y %I:%M:%S %p'
                    )


class CordClient(object):
    """
    Cord client. Allows you to query db items associated
    with a project (e.g. label rows, datasets).
    """

    def __init__(self, querier: Querier, config: Config):
        self._querier = querier
        self._config: Config = config

    @staticmethod
    def initialise(resource_id=None, api_key=None) -> Union[CordClientProject, CordClientDataset]:
        """
        Create and initialize a Cord client from a resource ID and API key.

        Args:
            resource_id: either of
                - A project ID string.
                  If None, uses the CORD_PROJECT_ID environment variable.
                - A dataset ID string.
                  If None, uses the CORD_DATASET_ID environment variable.
            api_key: An API key.
                     If None, uses the CORD_API_KEY environment variable.

        Returns:
            CordClient: A Cord client instance.
        """
        config = CordConfig(resource_id, api_key)
        return CordClient.initialise_with_config(config)

    @staticmethod
    def initialise_with_config(config: Config) -> Union[CordClientProject, CordClientDataset]:
        """
        Create and initialize a Cord client from a Cord config instance.

        Args:
            config: A Cord config instance.

        Returns:
            CordClient: A Cord client instance.
        """
        querier = Querier(config)
        key_type = querier.basic_getter(ApiKeyMeta)
        resource_type = key_type.get('resource_type', None)

        if resource_type == TYPE_PROJECT:
            logging.info("Initialising Cord client for project using key: %s",
                         key_type.get('title', ''))
            return CordClientProject(querier, config)

        elif resource_type == TYPE_DATASET:
            logging.info("Initialising Cord client for dataset using key: %s",
                         key_type.get('title', ''))
            return CordClientDataset(querier, config)

        else:
            raise cord.exceptions.InitialisationError(
                message=f"API key [{config.api_key}] is not associated with a project or dataset"
            )

    def __getattr__(self, name):
        """Overriding __getattr__."""
        value = self.__dict__.get(name)
        if not value:
            self_type = type(self).__name__
            if (self_type == "CordClientDataset" and
                    name in CordClientProject.__dict__.keys()):
                raise cord.exceptions.CordException(
                    message=('{} is implemented in Projects, not Datasets.'
                             .format(name))
                )
            elif (self_type == "CordClientProject" and
                  name in CordClientDataset.__dict__.keys()):
                raise cord.exceptions.CordException(
                    message=('{} is implemented in Datasets, not Projects.'
                             .format(name))
                )
            elif name == 'items':
                pass
            else:
                raise cord.exceptions.CordException(
                    message='{} is not implemented.'.format(name)
                )
        return value

    def get_cloud_integrations(self) -> List[CloudIntegration]:
        return self._querier.get_multiple(CloudIntegration)


class CordClientDataset(CordClient):
    def get_dataset(self):
        """
        Retrieve dataset info (pointers to data, labels).

        Args:
            self: Cord client object.

        Returns:
            Dataset: A dataset record instance.

        Raises:
            AuthorisationError: If the dataset API key is invalid.
            ResourceNotFoundError: If no dataset exists by the specified dataset ID.
            UnknownError: If an error occurs while retrieving the dataset.
        """
        return self._querier.basic_getter(Dataset)

    def upload_video(self, file_path):
        """
        Upload video to Cord storage.

        Args:
            self: Cord client object.
            file_path: path to video e.g. '/home/user/data/video.mp4'

        Returns:
            Bool.

        Raises:
            UploadOperationNotSupportedError: If trying to upload to external
                                              datasets (e.g. S3/GPC/Azure)
        """
        if os.path.exists(file_path):
            short_name = os.path.basename(file_path)
            signed_url = self._querier.basic_getter(
                SignedVideoURL,
                uid=short_name
            )
            res = upload_to_signed_url(
                file_path,
                signed_url,
                self._querier,
                Video
            )
            if res:
                logging.info("Upload complete.")
                logging.info("Please run client.get_dataset() to refresh.")
                return res
            else:
                raise cord.exceptions.CordException(
                    message='An error has occurred during video upload.'
                )
        else:
            raise cord.exceptions.CordException(
                message='{} does not point to a file.'.format(file_path)
            )

    def create_image_group(self, file_paths):
        """
        Create an image group in Cord storage.

        Args:
            self: Cord client object.
            file_paths: a list of paths to images, e.g.
                ['/home/user/data/img1.png', '/home/user/data/img2.png']

        Returns:
            Bool.

        Raises:
            UploadOperationNotSupportedError: If trying to upload to external
                                              datasets (e.g. S3/GPC/Azure)
        """
        for file_path in file_paths:
            if not os.path.exists(file_path):
                raise cord.exceptions.CordException(
                    message='{} does not point to a file.'.format(file_path)
                )
        short_names = list(map(os.path.basename, file_paths))
        signed_urls = self._querier.basic_getter(
            SignedImagesURL,
            uid=short_names
        )
        image_hash_list = upload_to_signed_url_list(
            file_paths, signed_urls, self._querier, Image
        )
        res = self._querier.basic_setter(
            ImageGroup,
            uid=image_hash_list,
            payload={}
        )
        if res:
            titles = [video_data.get('title') for video_data in res]
            logging.info("Upload successful! {} created.".format(titles))
            logging.info("Please run client.get_dataset() to refresh.")
            return res
        else:
            raise cord.exceptions.CordException(
                message='An error has occurred during image group creation.'
            )

    def delete_image_group(self, data_hash: str):
        """
        Create an image group in Cord storage.

        Args:
            self: Cord client object.
            data_hash: the hash of the image group you'd like to delete
        """
        self._querier.basic_delete(
            ImageGroup,
            uid=data_hash
        )

    def delete_data(self, data_hashes: list):
        """
        Delete a video/image group from a dataset.

        Args:
            self: Cord client object.
            data_hashes: list of hash of the videos/image_groups you'd like to delete, all should belong to the same
             dataset
        """
        self._querier.basic_delete(
            Video,
            uid=data_hashes
        )

    def add_private_data_to_dataset(self,
                                    integration_id: str,
                                    private_files: Union[str, typing.Dict, Path, typing.TextIO],
                                    ignore_errors: bool = False):
        """
        Append data hosted on private clouds to existing dataset

        Args:
            integration_id: str
                ID of the cloud integration to be used when accessing those files
            private_files:
                A str path to a file, Path object, json string or python dictionary of the files you wish to add
            ignore_errors: bool, optional
                Ignore individual errors when trying to access the specified files
        """
        if isinstance(private_files, dict):
            files = private_files
        elif isinstance(private_files, str):
            if os.path.exists(private_files):
                text_contents = Path(private_files).read_text()
            else:
                text_contents = private_files

            files = json.loads(text_contents)
        elif isinstance(private_files, Path):
            text_contents = private_files.read_text()
            files = json.loads(text_contents)
        elif isinstance(private_files, typing.TextIO):
            text_contents = private_files.read()
            files = json.loads(text_contents)
        else:
            raise ValueError(f'Type [{type(private_files)}] of argument private_files is not supported')

        payload = {
            'files': files,
            'integration_id': integration_id,
            'ignore_errors': ignore_errors
        }

        self._querier.basic_setter(DatasetData, self._config.resource_id, payload=payload)

    def re_encode_data(self, data_hashes: list):
        """
        Lanches a async task that can re-encode a list of videos.

        Args:
            self: Cord client object.
            data_hashes: list of hash of the videos you'd like to re_encode, all should belong to the same
             dataset
        Returns:
            ID(integer) of the async task launched.

        """
        payload = {"data_hash": data_hashes}
        return self._querier.basic_put(
            ReEncodeVideoTask,
            uid=None,
            payload=payload
        )

    def re_encode_data_status(self, job_id: int):
        """
        Returns the status of an existing async task which is aimed at re-encoding videos.

        Args:
            self: Cord client object.
            job_id: id of the async task that was launched to re-encode the videos

        Returns:
            ReEncodeVideoTask: Object containing the status of the task, along with info about the new encoded videos
             in case the task has been completed
        """
        return self._querier.basic_getter(
            ReEncodeVideoTask,
            uid=job_id
        )


class CordClientProject(CordClient):
    def get_project(self):
        """
        Retrieve project info (pointers to data, labels).

        Args:
            self: Cord client object.

        Returns:
            Project: A project record instance.

        Raises:
            AuthorisationError: If the project API key is invalid.
            ResourceNotFoundError: If no project exists by the specified project ID.
            UnknownError: If an error occurs while retrieving the project.
        """
        return self._querier.basic_getter(Project)

    def add_users(self, user_emails: List[str], user_role: ProjectUserRole) -> List[ProjectUser]:
        """
        Add users to project

        Args:
            user_emails: list of user emails to be added
            user_role: the user role to assign to all users

        Returns:
            ProjectUser

        Raises:
            AuthorisationError: If the project API key is invalid.
            ResourceNotFoundError: If no project exists by the specified project ID.
            UnknownError: If an error occurs while adding the users to the project
        """

        payload = {
            "user_emails": user_emails,
            "user_role": user_role
        }

        users = self._querier.basic_setter(ProjectUsers, self._config.resource_id, payload=payload)

        return list(map(lambda user: ProjectUser.from_dict(user), users))

    def copy_project(self, copy_labels=False, copy_models=False):
        """
        Copy the current project into a new one with copied contents including settings, datasets and users.
        Labels and models are optional
        Args:
            copy_labels: currently if labels is True, all tasks with labelling will be marked as complete,
                    otherwise all tasks will be recreated anew
            copy_models: currently if True, all models with their training information will be copied into the new
                         project

        Returns:
            bool

        Raises:
            AuthorisationError: If the project API key is invalid.
            ResourceNotFoundError: If no project exists by the specified project ID.
            UnknownError: If an error occurs while copying the project.
        """

        payload = {"copy_project_options": []}
        if copy_labels:
            payload["copy_project_options"].append("labels")
        if copy_models:
            payload["copy_project_options"].append("models")
        return self._querier.basic_setter(ProjectCopy, self._config.resource_id, payload=payload)

    def get_label_row(self, uid: str, get_signed_url: bool = True):
        """
        Retrieve label row.

        Args:
            uid: A label_hash (uid) string.
            get_signed_url: By default the operation returns a signed URL for the underlying data asset. This can be
            expensive so it can optionally be turned off

        Returns:
            LabelRow: A label row instance.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            ResourceNotFoundError: If no label exists by the specified label_hash (uid).
            UnknownError: If an error occurs while retrieving the label.
            OperationNotAllowed: If the read operation is not allowed by the API key.
        """
        payload = {
            'get_signed_url': get_signed_url
        }

        return self._querier.basic_getter(LabelRow, uid, payload=payload)

    def save_label_row(self, uid, label):
        """
        Save existing label row.

        If you have a series of frame labels and have not updated answer
        dictionaries, call the construct_answer_dictionaries utilities function
        to do so prior to saving labels.

        Args:
            uid: A label_hash (uid) string.
            label: A label row instance.

        Returns:
            Bool.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            ResourceNotFoundError: If no label exists by the specified label_hash (uid).
            UnknownError: If an error occurs while saving the label.
            OperationNotAllowed: If the write operation is not allowed by the API key.
            AnswerDictionaryError: If an object or classification instance is missing in answer dictionaries.
            CorruptedLabelError: If a blurb is corrupted (e.g. if the frame labels have more frames than the video).
        """
        label = LabelRow(label)
        return self._querier.basic_setter(LabelRow, uid, payload=label)

    def create_label_row(self, uid):
        """
        Create a label row (for data in a project not previously been labeled).

        Args:
            uid: the data_hash (uid) of the data unit being labeled.
                Available in client.get_project().get('label_rows')
                where label_status is NOT_LABELLED.

        Returns:
            LabelRow: A label row instance.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            UnknownError: If an error occurs while saving the label.
            OperationNotAllowed: If the write operation is not allowed by the API key.
            AnswerDictionaryError: If an object or classification instance is missing in answer dictionaries.
            CorruptedLabelError: If a blurb is corrupted (e.g. if the frame labels have more frames than the video).
            ResourceExistsError: If a label row already exists for this project data. Avoids overriding existing work.
        """
        return self._querier.basic_put(LabelRow, uid=uid, payload=None)

    def add_datasets(self, dataset_hashes: List[str]):
        """
        Add a dataset to a project

        Args:
            dataset_hashes: List of dataset hashes of the datasets to be added

        Returns:
            Bool.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            ResourceNotFoundError: If one or more datasets don't exist by the specified dataset_hashes.
            UnknownError: If an error occurs while adding the datasets to the project.
            OperationNotAllowed: If the write operation is not allowed by the API key.
        """
        payload = {"dataset_hashes": dataset_hashes}
        return self._querier.basic_setter(ProjectDataset, uid=None, payload=payload)

    def remove_datasets(self, dataset_hashes: List[str]):
        """
        Remove datasets from project

        Args:
            dataset_hashes: List of dataset hashes of the datasets to be removed

        Returns:
            Bool.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            ResourceNotFoundError: If no dataset exists by the specified dataset_hash (uid).
            UnknownError: If an error occurs while removing the datasets from the project.
            OperationNotAllowed: If the operation is not allowed by the API key.
        """
        return self._querier.basic_delete(ProjectDataset, uid=dataset_hashes)

    def get_project_ontology(self) -> Ontology:
        project = self.get_project()
        ontology = project['editor_ontology']
        return Ontology.from_dict(ontology)

    def add_object(self, name: str, shape: ObjectShape) -> bool:
        """
        Add object to an ontology
        Args:
            name: the name of the object
            shape: the shape of the object (BOUNDING_BOX, POLYGON or KEY_POINT)

        Returns:
            bool

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            UnknownError: If an error occurs while add te object to the project ontology
            OperationNotAllowed: If the operation is not allowed by the API key.
        """
        ontology = self.get_project_ontology()
        ontology.add_object(name, shape)
        return self.__set_project_ontology(ontology)

    def add_classification(self, name: str, classification_type: ClassificationType, required: bool,
                           options: Optional[typing.Iterable[str]] = None):
        """

        Args:
            name: the name of the classification
            classification_type: the classification type (RADIO, TEXT or CHECKLIST)
            required (whether this classification is required by the annotator):
            options: the list of options for the classification (to be set to None for texts)

        Returns:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            UnknownError: If an error occurs while add te classification to the project ontology
            OperationNotAllowed: If the operation is not allowed by the API key.
        """

        ontology = self.get_project_ontology()
        ontology.add_classification(name, classification_type, required, options)
        return self.__set_project_ontology(ontology)


    def create_model_row(self,
                         title=None,
                         description=None,
                         features=None,
                         model=None,
                         ):
        """
        Create a model row.

        Args:
            title: Model title.
            description: Model description.
            features: List of feature feature uid's (hashes) to be included in the model.
            model: Model (resnet18, resnet34, resnet50,
                resnet101, resnet152, vgg16, vgg19, yolov5, faster_rcnn, mask_rcnn).

        Returns:
            ModelRow: A model_hash (uid) string.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            ModelFeaturesInconsistentError: If a feature type is different than what is supported by the model (e.g. if
                creating a classification model using a bounding box).
        """
        if title is None:
            raise cord.exceptions.CordException(
                message='You must set a title to create a model row.'
            )

        if features is None:
            raise cord.exceptions.CordException(
                message="You must pass a list of feature uid's (hashes) to create a model row."
            )

        if model is None or model not in [RESNET18, RESNET34, RESNET50, RESNET101, RESNET152, VGG16, VGG19, YOLOV5,
                                          FASTER_RCNN, MASK_RCNN]:
            raise cord.exceptions.CordException(
                message="You must pass a model (resnet18, resnet34, resnet50, resnet101, resnet152, vgg16, vgg19, "
                        "yolov5, faster_rcnn, mask_rcnn) to create a model row."
            )

        model_row = ModelRow({
            'title': title,
            'description': description,
            'features': features,
            'model': model,
        })

        model = Model({
            'model_operation': ModelOperations.CREATE.value,
            'model_parameters': model_row,
        })

        return self._querier.basic_put(Model, None, payload=model)

    def model_inference(self,
                        uid,
                        file_paths=None,
                        base64_strings=None,
                        conf_thresh=0.6,
                        iou_thresh=0.3,
                        device="cuda",
                        detection_frame_range=None,
                        allocation_enabled=False,
                        ):
        """
        Run inference with model trained on the platform.

        Args:
            uid: A model_iteration_hash (uid) string.
            file_paths: List of local file paths to image(s) or video(s) - if running inference on files.
            base64_strings: List of base 64 strings of image(s) or video(s) - if running inference on base64 strings.
            conf_thresh: Confidence threshold (default 0.6).
            iou_thresh: Intersection over union threshold (default 0.3).
            device: Device (CPU or CUDA, default is CUDA).
            detection_frame_range: Detection frame range (for videos).
            allocation_enabled: Object UID allocation (tracking) enabled (disabled by default).

        Returns:
            Inference results: A dict of inference results.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            ResourceNotFoundError: If no model exists by the specified model_iteration_hash (uid).
            UnknownError: If an error occurs while running inference.
            FileTypeNotSupportedError: If the file type is not supported for inference (has to be an image or video)
            DetectionRangeInvalidError: If a detection range is invalid for video inference
        """
        if (file_paths is None and base64_strings is None) or (
                file_paths is not None and len(file_paths) > 0 and base64_strings is not None and len(
                base64_strings) > 0):
            raise cord.exceptions.CordException(
                message='To run model inference, you must pass either a list of files or base64 strings.'
            )

        if detection_frame_range is None:
            detection_frame_range = []

        files = []
        if file_paths is not None:
            for file_path in file_paths:
                file = open(file_path, 'rb').read()
                files.append({
                    "uid": file_path,  # Add file path as inference identifier
                    "base64_str": base64.b64encode(file).decode('utf-8')  # File to base64 string
                })

        elif base64_strings is not None:
            for base64_string in base64_strings:
                files.append({
                    "uid": str(uuid.uuid4()),  # Add uuid as inference identifier
                    "base64_str": base64_string.decode('utf-8')  # base64 string to utf-8
                })

        inference_params = ModelInferenceParams({
            'files': files,
            'conf_thresh': conf_thresh,
            'iou_thresh': iou_thresh,
            'device': device,
            'detection_frame_range': detection_frame_range,
            'allocation_enabled': allocation_enabled,
        })

        model = Model({
            'model_operation': ModelOperations.INFERENCE.value,
            'model_parameters': inference_params,
        })

        return self._querier.basic_setter(Model, uid, payload=model)

    def model_train(self,
                    uid,
                    label_rows=None,
                    epochs=None,
                    batch_size=24,
                    weights=None,
                    device="cuda"
                    ):
        """
        Train a model created on the platform.

        Args:
            uid: A model_hash (uid) string.
            label_rows: List of label row uid's (hashes) for training.
            epochs: Number of passes through training dataset.
            batch_size: Number of training examples utilized in one iteration.
            weights: Model weights.
            device: Device (CPU or CUDA, default is CUDA).

        Returns:
            A model iteration object.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            ModelWeightsInconsistentError: If the passed model weights are incompatible with the selected model.
            ResourceNotFoundError: If no model exists by the specified model_hash (uid).
            UnknownError: If an error occurs during training.
        """
        if label_rows is None:
            raise cord.exceptions.CordException(
                message="You must pass a list of label row uid's (hashes) to train a model."
            )

        if epochs is None:
            raise cord.exceptions.CordException(
                message="You must set number of epochs to train a model."
            )

        if batch_size is None:
            raise cord.exceptions.CordException(
                message="You must set a batch size to train a model."
            )

        if weights is None:
            raise cord.exceptions.CordException(
                message="You must select model weights to train a model."
            )

        if device is None:
            raise cord.exceptions.CordException(
                message="You must set a device (cuda or CPU) train a model."
            )

        training_params = ModelTrainingParams({
            'label_rows': label_rows,
            'epochs': epochs,
            'batch_size': batch_size,
            'weights': weights,
            'device': device,
        })

        model = Model({
            'model_operation': ModelOperations.TRAIN.value,
            'model_parameters': training_params,
        })

        return self._querier.basic_setter(Model, uid, payload=model)

    def object_interpolation(self,
                             key_frames,
                             objects_to_interpolate,
                             ):
        """
        Run object interpolation algorithm on project labels (requires an editor ontology and feature uid's).

        Interpolation is supported for bounding box, polygon, and keypoint.

        Args:
            key_frames: Labels for frames to be interpolated. Key frames are consumed in the form:

                "frame": {
                    "objects": [
                        {
                            "objectHash": object_uid,
                            "featureHash": feature_uid (from editor ontology),
                            "polygon": {
                                "0": {
                                    "x": x1,
                                    "y": y1,
                                },
                                "1": {
                                    "x": x2,
                                    "y": y2,
                                },
                                "2" {
                                    "x": x3,
                                    "y": y3,
                                },
                                ...,
                            }
                        },
                        {
                            ...
                        }
                    ]
                },
                "frame": {
                    ...,
                }

            objects_to_interpolate: List of object uid's (hashes) of objects to interpolate.

        Returns:
            Interpolation results: Full set of filled frames including interpolated objects.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            UnknownError: If an error occurs while running interpolation.
        """
        if len(key_frames) == 0 or len(objects_to_interpolate) == 0:
            raise cord.exceptions.CordException(
                message="To run object interpolation, you must pass key frames and objects to interpolate.."
            )

        interpolation_params = ObjectInterpolationParams({
            'key_frames': key_frames,
            'objects_to_interpolate': objects_to_interpolate,
        })

        algo = LabelingAlgorithm({
            'algorithm_name': INTERPOLATION,
            'algorithm_parameters': interpolation_params,
        })

        return self._querier.basic_setter(LabelingAlgorithm, str(uuid.uuid4()), payload=algo)

    def fitted_bounding_boxes(self,
                              frames: dict,
                              video: dict,
                              ):
        """

        Args:
            frames: Labels for frames to be fitted. Frames are consumed in the form:

                "frame": {
                    "objects": [
                        {
                            "objectHash": object_uid,
                            "featureHash": feature_uid (from editor ontology),
                            "polygon": {
                                "0": {
                                    "x": x1,
                                    "y": y1,
                                },
                                "1": {
                                    "x": x2,
                                    "y": y2,
                                },
                                "2" {
                                    "x": x3,
                                    "y": y3,
                                },
                                ...,
                            }
                        },
                        {
                            ...
                        }
                    ]
                },
                "frame": {
                    ...,
                }

            video: Metadata of the video for which bounding box fitting needs to be run

                {
                    "width" : w,
                    "height" : h,
                }

        Returns:
            Fitting results: Full set of filled frames including fitted objects.

        Raises:
            AuthenticationError: If the project API key is invalid.
            AuthorisationError: If access to the specified resource is restricted.
            UnknownError: If an error occurs while running interpolation.
        """
        if len(frames) == 0 or len(video) == 0:
            raise cord.exceptions.CordException(
                message="To run fitting, you must pass frames and video to run bounding box fitting on.."
            )

        fitting_params = BoundingBoxFittingParams({
            'labels': frames,
            'video': video,
        })

        algo = LabelingAlgorithm({
            'algorithm_name': FITTED_BOUNDING_BOX,
            'algorithm_parameters': fitting_params,
        })

        return self._querier.basic_setter(LabelingAlgorithm, str(uuid.uuid4()), payload=algo)

    def get_data(self, data_hash: str, get_signed_url: bool = False) -> Tuple[Optional[Video], Optional[List[Image]]]:
        """
                Retrieve information about a video or image group.

                Params:
                    data_hash: The uid of the data object
                    get_signed_url: Optionally return signed URLs for timed public access to that resource
                                    (default False)

                Returns:
                    A consisting of the video (if it exists) and a list of individual images (if they exist)

                Raises:
                    AuthenticationError: If the project API key is invalid.
                    AuthorisationError: If access to the specified resource is restricted.
                    UnknownError: If an error occurs while retrieving the object.
                """
        uid = {'data_hash': data_hash,
               'get_signed_url': get_signed_url}

        dataset_data: DatasetData = self._querier.basic_getter(DatasetData, uid=uid)

        video: Union[Video, None] = None
        if dataset_data['video'] is not None:
            video = Video(dataset_data['video'])

        images: Union[List[Image], None] = None
        if dataset_data['images'] is not None:
            images = []

            for image in dataset_data['images']:
                images.append(Image(image))

        return video, images

    def get_websocket_url(self) -> str:
        return f'{self._config.websocket_endpoint}?' \
               f'client_type={2}&' \
               f'project_hash={self._config.resource_id}&' \
               f'api_key={self._config.api_key}'

    def get_label_logs(self, user_hash: str = None, data_hash: str = None, from_unix_seconds: int = None,
                       to_unix_seconds: int = None) -> List[LabelLog]:

        function_arguments = locals()

        query_payload = {
            k: v for (k, v) in function_arguments.items() if k is not 'self' and v is not None
        }

        return self._querier.get_multiple(LabelLog, payload=query_payload)

    def __set_project_ontology(self, ontology: Ontology) -> bool:

        """
        Save updated project ontology
        Args:
            ontology: the updated project ontology

        Returns:
            bool
        """
        payload = {"editor": ontology.to_dict()}
        return self._querier.basic_setter(Project, uid=None, payload=payload)



import uuid
import pytest

import cord.exceptions
from cord.client import CordClient
from cord.configs import CordConfig
from cord.orm.project import Project
from cord.orm.label_row import LabelRow

from tests.test_data.interpolation_test_blurb import INTERPOLATION_TEST_BLURB
from tests.test_data.test_blurb import TEST_BLURB
from tests.test_data.img_group_test_blurb import IMG_GROUP_TEST_BLURB

# Dummy keys, can be used and abused
LABEL_READ_WRITE_KEY = 'Igr3RTx7B4gJbHZM0eyjOXaPr7jg22Fw22AQbYT0nQM'
LABEL_READ_KEY = 'aEfra0xAWVOsNzW7sXhGMIPzuh5FVfflS0DiuB648Ak'
LABEL_WRITE_KEY = 'cWNtJAzzlw3eBWDTMrDPJy9iAXn9eJ0sP8yRj3EVi1U'


@pytest.fixture
def keys():
    resource_id = 'dd00ab81-0834-481b-9ef5-49e35f9f7b63'  # Dummy project ID
    key = LABEL_READ_WRITE_KEY  # Dummy API key with label read/write access
    label_id = '6786fa5a-3b48-4d34-a7c5-ed2ff82bd3ba'  # Dummy video label row ID
    img_group_label_id = '5fbba385-4918-4eee-85a8-8b7a2e71dd16'  # Image group label row ID
    return resource_id, key, label_id, img_group_label_id


@pytest.fixture
def client(keys):
    return CordClient.initialise(resource_id=keys[0], api_key=keys[1])


def test_initialise(keys):
    assert isinstance(CordClient.initialise(resource_id=keys[0], api_key=keys[1]), CordClient)


def test_initialise_with_config(keys):
    config = CordConfig(resource_id=keys[0], api_key=keys[1])
    assert isinstance(CordClient.initialise_with_config(config), CordClient)


def test_missing_key(keys):
    with pytest.raises(expected_exception=cord.exceptions.AuthenticationError) as excinfo:
        CordClient.initialise(resource_id=keys[0])

    assert excinfo.value.message == "API key not provided"


def test_missing_resource_id(keys):
    with pytest.raises(expected_exception=cord.exceptions.AuthenticationError) as excinfo:
        CordClient.initialise(api_key=keys[1])

    assert excinfo.value.message == "Project ID or dataset ID not provided"


def test_invalid_key(keys):
    with pytest.raises(expected_exception=cord.exceptions.AuthenticationError):
        CordClient.initialise(keys[0], uuid.uuid4())


def test_invalid_resource_id(keys):
    with pytest.raises(expected_exception=cord.exceptions.AuthenticationError):
        CordClient.initialise(uuid.uuid4(), keys[1])


def test_get_project(client):
    assert isinstance(client.get_project(), Project)


def test_get_label_blurb(keys, client):
    assert isinstance(client.get_label_row(keys[2]), LabelRow)


def test_get_label_with_invalid_id_throws_authorisation_exception(client):
    with pytest.raises(expected_exception=cord.exceptions.AuthorisationError):
        client.get_label_row('test')


def test_get_label_with_write_key_throws_operation_not_allowed_exception(keys):
    client = CordClient.initialise(keys[0], LABEL_WRITE_KEY)

    with pytest.raises(expected_exception=cord.exceptions.OperationNotAllowed):
        client.get_label_row(keys[2])


def test_save_video_label_row(keys, client):
    blurb = client.save_label_row(keys[2], TEST_BLURB)
    assert blurb is True


def test_save_img_group_label_row(keys, client):
    blurb = client.save_label_row(keys[3], IMG_GROUP_TEST_BLURB)
    assert blurb is True


def test_save_label_with_invalid_id_throws_authorisation_exception(keys, client):
    with pytest.raises(expected_exception=cord.exceptions.AuthorisationError):
        client.save_label_row('test', TEST_BLURB)


def test_save_label_with_read_key_throws_operation_not_allowed_exception(keys):
    client = CordClient.initialise(keys[0], LABEL_READ_KEY)

    with pytest.raises(expected_exception=cord.exceptions.OperationNotAllowed):
        client.save_label_row(keys[2], TEST_BLURB)


def test_object_interpolation_with_polygons(keys):
    client = CordClient.initialise(keys[0], LABEL_READ_KEY)
    objects = client.object_interpolation(INTERPOLATION_TEST_BLURB, ['60f75ddb-aa68-4654-8c85-f6959dbb62eb'])
    assert isinstance(objects, dict)





import uuid
from typing import List, Dict, Optional, Iterable

from cord.project_ontology.classification_attribute import ClassificationAttribute
from cord.project_ontology.classification_option import ClassificationOption
from cord.project_ontology.classification_type import ClassificationType
from cord.project_ontology.object_type import ObjectShape
from cord.project_ontology.ontology_classification import OntologyClassification
from cord.project_ontology.ontology_object import OntologyObject


# removal may require calling label/delete/object and label/delete/attribute
def generate_feature_node_hash() -> str:
    return str(uuid.uuid4())[0:8]


class Ontology:
    COLORS = (
        '#D33115',
        '#E27300',
        '#16406C',
        '#FE9200',
        '#FCDC00',
        '#DBDF00',
        '#A4DD00',
        '#68CCCA',
        '#73D8FF',
        '#AEA1FF',
        '#FCC400',
        '#B0BC00',
        '#68BC00',
        '#16A5A5',
        '#009CE0',
        '#7B64FF',
        '#FA28FF',
        '#B3B3B3',
        '#9F0500',
        '#C45100',
        '#FB9E00',
        '#808900',
        '#194D33',
        '#0C797D',
        '#0062B1',
        '#653294',
        '#AB149E',
    )

    def __init__(self):
        self.ontology_objects: List[OntologyObject] = []
        self.ontology_classifications: List[OntologyClassification] = []
        self.color_index = 0

    def __str__(self):
        return str(self.to_dict())

    def __current_object_id(self):
        return len(self.ontology_objects) + 1

    def __current_classification_id(self):
        return len(self.ontology_classifications) + 1

    @classmethod
    def from_dict(cls, ontology_dict: Dict):
        ontology = Ontology()
        ontology.ontology_objects = cls.__build_ontology_objects(ontology_dict['objects'])
        ontology.ontology_classifications = cls.__build_ontology_classifications(ontology_dict['classifications'])
        ontology.color_index = len(ontology.ontology_objects) % len(cls.COLORS)
        return ontology

    @classmethod
    def __build_ontology_objects(cls, ontology_objects_dict: List[Dict]):
        ontology_objects = []
        for ontology_object_dict in ontology_objects_dict:
            ontology_object = OntologyObject(ontology_object_dict['id'],
                                             ontology_object_dict['color'],
                                             ontology_object_dict['name'],
                                             ObjectShape(ontology_object_dict['shape']),
                                             ontology_object_dict['featureNodeHash'])
            ontology_objects.append(ontology_object)
        return ontology_objects

    @classmethod
    def __build_ontology_classifications(cls, ontology_classifications_dict: List[Dict]):
        ontology_classifications = []
        for classification_object_dict in ontology_classifications_dict:
            classification_attributes = []
            for attribute in classification_object_dict['attributes']:
                classification_attribute = ClassificationAttribute(attribute['id'],
                                                                   attribute['name'],
                                                                   ClassificationType(attribute['type']),
                                                                   attribute['required'],
                                                                   attribute['featureNodeHash'],
                                                                   )
                options_dict = attribute.get('options')
                if options_dict is not None:
                    options = cls.__build_classification_options(attribute['options'])
                    classification_attribute.options = options
                classification_attributes.append(classification_attribute)
            ontology_classification = OntologyClassification(classification_object_dict['id'],
                                                             classification_object_dict['featureNodeHash'],
                                                             classification_attributes)
            ontology_classifications.append(ontology_classification)
        return ontology_classifications

    @classmethod
    def __build_classification_options(cls, options_dict: List[Dict]):
        options = []
        for option_dict in options_dict:
            option = ClassificationOption(option_dict['id'],
                                          option_dict['label'],
                                          option_dict['value'],
                                          option_dict['featureNodeHash'])
            options.append(option)
        return options

    def add_object(self, name: str, shape: ObjectShape) -> None:
        ontology_object = OntologyObject(str(self.__current_object_id()),
                                         self.COLORS[self.color_index],
                                         name,
                                         shape,
                                         generate_feature_node_hash())
        self.ontology_objects.append(ontology_object)
        self.color_index = (self.color_index + 1) % len(self.COLORS)

    def add_classification(self, name: str, classification_type: ClassificationType, required: bool,
                           options: Optional[Iterable[str]] = None) -> None:
        ontology_classification = OntologyClassification(str(self.__current_classification_id()),
                                                         generate_feature_node_hash(),
                                                         [self.__create_classification_attributes(name,
                                                                                                  classification_type,
                                                                                                  required,
                                                                                                  options)])
        self.ontology_classifications.append(ontology_classification)

    def __create_classification_attributes(self,
                                           name: str,
                                           classification_type: ClassificationType,
                                           required: bool,
                                           options: Optional[Iterable[str]]) -> ClassificationAttribute:
        attribute_id = 1
        classification_attribute = ClassificationAttribute(f"{self.__current_classification_id()}.{attribute_id}",
                                                           name,
                                                           classification_type,
                                                           required,
                                                           generate_feature_node_hash())
        if options:
            classification_options = []
            for option_id, option in enumerate(options, 1):
                classification_option = ClassificationOption(
                    f"{self.__current_classification_id()}.{attribute_id}.{option_id}",
                    option,
                    self.__format_option_value(option),
                    generate_feature_node_hash())
                classification_options.append(classification_option)
            classification_attribute.options = classification_options

        return classification_attribute

    def to_dict(self) -> dict:
        objects = []
        classifications = []

        for ontology_object in self.ontology_objects:
            objects.append(self.ontology_object_to_dict(ontology_object))

        for ontology_classification in self.ontology_classifications:
            classifications.append(self.ontology_classification_to_dict(ontology_classification))

        return {'objects': objects, 'classifications': classifications}

    def ontology_object_to_dict(self, ontology_object: OntologyObject) -> Dict:
        return {
            'id': ontology_object.id,
            'color': ontology_object.color,
            'name': ontology_object.name,
            'shape': ontology_object.shape.value,
            'featureNodeHash': ontology_object.feature_node_hash
        }

    def ontology_classification_to_dict(self, ontology_classification: OntologyClassification) -> Dict:
        classification = {
            'id': ontology_classification.id,
            'featureNodeHash': ontology_classification.feature_node_hash,
        }
        attributes = []
        for classification_attribute in ontology_classification.attributes:
            attribute = {
                'id': classification_attribute.id,
                'name': classification_attribute.name,
                'type': classification_attribute.classification_type.value,
                'required': classification_attribute.required,
                'featureNodeHash': classification_attribute.feature_node_hash
            }
            if classification_attribute.options:
                options = []
                for classification_option in classification_attribute.options:
                    option = {
                        'id': classification_option.id,
                        'label': classification_option.label,
                        'value': classification_option.value,
                        'featureNodeHash': classification_option.feature_node_hash
                    }
                    options.append(option)
                attribute['options'] = options
            attributes.append(attribute)

        classification['attributes'] = attributes
        return classification

    def __format_option_value(self, label: str):
        return label.replace(" ", "_")



import os
import unittest
import uuid
from datetime import datetime, timezone

from cord.client import CordClient
from cord.exceptions import (
    AuthenticationError, AuthorisationError,
    OperationNotAllowed, ResourceExistsError)
from cord.orm.label_row import LabelRow
from tests.test_data.interpolation_test_blurb import INTERPOLATION_TEST_BLURB
from tests.test_data.test_blurb import TEST_BLURB

LABEL_READ_WRITE_KEY = 'aJpUqf1UifGGc-c10MjbQ_ze7G6SOs0d-SwfO8D_Ntg'
LABEL_READ_KEY = '1xI0y7cHYPPZRzCpDvk9WIBMfeonkKVBxjjgJHuYiUI'
LABEL_WRITE_KEY = 'CPNuaX5_eubBdx9CJ-sZrj1ZkqL_51L-Dd27dZ9bi_4'
PROJECT_ID = '5036161d-3da4-4711-8301-d1901fc7b349'
DATA_ID = 'e8296743-577e-4b5b-b151-0a6254535ed3'
IMG_GROUP_DATA_ID = ''
LABEL_ID = '23813f86-b352-4c61-bde1-bf8ab00611d5'
OH = "0b402f38-bc7b-4cae-be94-f963a6713de5"

DATASET_ID = '93acfbf8-77db-48d9-90a9-2a9b9065511c'
DATASET_KEY = 'G26da4vZFI096OSnwSVZbO9TDIyYF2_OdODKNakJEQk'

"""
    Dev tests for Client (local)
"""


class UnitTests(unittest.TestCase):
    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        super(UnitTests, cls).setUpClass()
        cls.read_c = CordClient.initialise(PROJECT_ID, LABEL_READ_KEY)
        cls.write_c = CordClient.initialise(PROJECT_ID, LABEL_WRITE_KEY)
        cls.rw_c = CordClient.initialise(PROJECT_ID, LABEL_READ_WRITE_KEY)
        cls.dt_c = CordClient.initialise(DATASET_ID, DATASET_KEY)

    def test_create_label_wrong_data_hash(self):
        with self.assertRaises(AuthorisationError):
            self.rw_c.create_label_row('some_random_data_hash')

    def test_create_label_when_exists(self):
        with self.assertRaises(ResourceExistsError):
            self.rw_c.create_label_row(DATA_ID)
            self.rw_c.create_label_row(DATA_ID)

    def test_create_label_with_readonly(self):
        with self.assertRaises(OperationNotAllowed):
            self.read_c.create_label_row(DATA_ID)

    def test_1(self):
        with self.assertRaises(AuthenticationError) as excinfo:
            CordClient.initialise(PROJECT_ID)
        self.assertEqual(
            'API key not provided',
            str(excinfo.exception)
        )

    def test_2(self):
        with self.assertRaises(AuthenticationError) as excinfo:
            CordClient.initialise(api_key=LABEL_READ_WRITE_KEY)
        self.assertEqual(
            'Project ID or dataset ID not provided',
            str(excinfo.exception)
        )

    def test_3(self):
        with self.assertRaises(AuthenticationError):
            client = CordClient.initialise(PROJECT_ID, uuid.uuid4())

    def test_4(self):
        with self.assertRaises(AuthenticationError):
            client = CordClient.initialise(uuid.uuid4(), LABEL_READ_WRITE_KEY)

    def test_5(self):
        assert isinstance(self.rw_c.get_label_row(LABEL_ID), LabelRow)

    def test_6(self):
        with self.assertRaises(AuthorisationError):
            self.rw_c.get_label_row('test')

    def test_7(self):
        with self.assertRaises(OperationNotAllowed):
            self.write_c.get_label_row(LABEL_ID)

    def test_8(self):
        new_test_blurb = TEST_BLURB.copy()
        wrong_id = "97681cfa-0c3a-4ebe-aba3-f17e136fc231"
        new_test_blurb['data_units'][DATA_ID] = new_test_blurb['data_units'][wrong_id]
        del new_test_blurb['data_units'][wrong_id]
        new_test_blurb['data_units'][DATA_ID]["data_hash"] = DATA_ID
        blurb = self.rw_c.save_label_row(LABEL_ID, TEST_BLURB)
        assert blurb is True

    def test_10(self):
        with self.assertRaises(AuthorisationError):
            self.rw_c.save_label_row('test', TEST_BLURB)

    def test_11(self):
        with self.assertRaises(OperationNotAllowed):
            self.read_c.save_label_row(LABEL_ID, TEST_BLURB)

    def test_12(self):
        test_blurb = INTERPOLATION_TEST_BLURB.copy()
        for key, val in test_blurb.items():
            new_list = []
            for v in test_blurb[key]["objects"]:
                v["objectHash"] = OH
                new_list.append(v.copy())
            test_blurb[key]["objects"] = new_list.copy()
        objects = self.rw_c.object_interpolation(test_blurb,
                                                 ['60f75ddb-aa68-4654-8c85-f6959dbb62eb'])
        assert isinstance(objects, dict)

    def test_upload_video(self):
        path = os.path.dirname(__file__)
        video_path = os.path.join(path, 'test_data', 'media', 'video.mp4')
        self.dt_c.upload_video(video_path)
        now = datetime.now(timezone.utc).replace(tzinfo=None)
        dataset = self.dt_c.get_dataset()
        assert len(dataset) > 0
        passed = False
        for dt in dataset.get('data_rows', []):
            if dt.get('data_title', '') == 'video.mp4':
                then_time = datetime.strptime(
                    dt.get('created_at'), '%Y-%m-%d %H:%M:%S'
                )
                delta = int((now - then_time).total_seconds())
                if delta >= 0 & delta <= 1 * 60:
                    passed = True
        assert passed

    def test_upload_image_group(self):
        path = os.path.dirname(__file__)
        im1 = os.path.join(path, 'test_data', 'media', 'screen1.png')
        im2 = os.path.join(path, 'test_data', 'media', 'screen1.png')
        image_groups = self.dt_c.create_image_group([im1, im2])
        assert len(image_groups) == 1
        assert 'image-group-' in image_groups[0]['title'], 'Upload did not complete successfully'

    def test_upload_image_group_different_resolution(self):
        path = os.path.dirname(__file__)
        test_images_dir = os.path.join(path, 'test_data', 'media', 'upload_image_group_different_resolution')
        im1 = os.path.join(test_images_dir, 'car-1280-720.jpeg')
        im2 = os.path.join(test_images_dir, 'car-1920-1080.jpeg')

        image_groups = self.dt_c.create_image_group([im1, im2])
        assert len(image_groups) == 2
        for image_group in image_groups:
            assert 'image-group-' in image_group['title'], 'Upload did not complete successfully'


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import OrderedDict

from cord.orm import base_orm


class Project(base_orm.BaseORM):
    """
    A project defines a label ontology and is a collection of datasets and label rows.

    ORM:

    title,
    description,
    editor_ontology,
    datasets: [
       {
            dataset_hash (uid),
            title,
            description,
            dataset_type (internal vs. AWS/GCP/Azure),
       },
       ...
    ],
    label_rows: [
        {
            label_hash (uid),
            data_hash (uid),
            dataset_hash (uid),
            dataset_title,
            data_title,
            data_type,
            label_status
        },
        ...
    ]

    """

    DB_FIELDS = OrderedDict([
        ("title", str),
        ("description", str),
        ("editor_ontology", (dict, str)),
        ("datasets", (list, str)),
        ("label_rows", (list, str)),
    ])

    NON_UPDATABLE_FIELDS = {
        "editor_ontology",
        "datasets",
        "label_rows"
    }

    def get_labels_list(self):
        """
        Returns a list of all label row IDs (label_hash uid) in a project.
        """
        labels = self.to_dic().get('label_rows')
        res = []
        for label in labels:
            res.append(label.get('label_hash'))
        return res



class ProjectCopy:
    pass


class ProjectUsers:
    pass


class ProjectDataset:
    pass



from dataclasses import dataclass
from typing import List

from cord.project_ontology.classification_attribute import ClassificationAttribute


@dataclass
class OntologyClassification:
    id: str
    feature_node_hash: str
    attributes: List[ClassificationAttribute]




from dataclasses import dataclass
from typing import Optional, List, Iterable

from cord.project_ontology.classification_option import ClassificationOption
from cord.project_ontology.classification_type import ClassificationType


@dataclass
class ClassificationAttribute:
    id: str
    name: str
    classification_type: ClassificationType
    required: bool
    feature_node_hash: str
    options: Optional[Iterable[ClassificationOption]] = None

    def __setattr__(self, name, value):
        if (name == "classification_type" and value == ClassificationType.TEXT and self.__dict__.get("options")) or \
                (name == "options" and value and self.__dict__.get("classification_type") == ClassificationType.TEXT):
            raise Exception("cannot assign options to a classification text")
        self.__dict__[name] = value



from __future__ import annotations

from typing import List

from cord.configs import UserConfig
from cord.http.querier import Querier
from cord.orm.dataset import Dataset, DatasetType
from cord.orm.dataset import DatasetScope, DatasetAPIKey
from cord.orm.project import Project
from cord.orm.project_api_key import ProjectAPIKey
from cord.utilities.client_utilities import APIKeyScopes


class CordUserClient:
    def __init__(self, user_config: UserConfig, querier: Querier):
        self.user_config = user_config
        self.querier = querier

    def create_private_dataset(self, dataset_title: str, dataset_type: DatasetType, dataset_description: str = None):
        return self.create_dataset(dataset_title, dataset_type, dataset_description)

    def create_dataset(self, dataset_title: str, dataset_type: DatasetType, dataset_description: str = None):
        dataset = {
            'title': dataset_title,
            'type': dataset_type,
        }

        if dataset_description:
            dataset['description'] = dataset_description

        return self.querier.basic_setter(Dataset, uid=None, payload=dataset)

    def create_dataset_api_key(
            self, dataset_hash: str, api_key_title: str, dataset_scopes: List[DatasetScope]) -> DatasetAPIKey:
        api_key_payload = {
            'dataset_hash': dataset_hash,
            'title': api_key_title,
            'scopes': list(map(lambda scope: scope.value, dataset_scopes))
        }
        response = self.querier.basic_setter(DatasetAPIKey, uid=None, payload=api_key_payload)
        return DatasetAPIKey.from_dict(response)

    def get_dataset_api_keys(self, dataset_hash: str) -> List[DatasetAPIKey]:
        api_key_payload = {
            'dataset_hash': dataset_hash,
        }
        api_keys: List[DatasetAPIKey] = self.querier.get_multiple(DatasetAPIKey, uid=None, payload=api_key_payload)
        return api_keys

    @staticmethod
    def create_with_ssh_private_key(ssh_private_key: str, password: str = None, **kwargs) -> CordUserClient:
        user_config = UserConfig.from_ssh_private_key(ssh_private_key, password, **kwargs)
        querier = Querier(user_config)

        return CordUserClient(user_config, querier)

    def create_project(self, project_title: str, dataset_hashes: List[str], project_description: str = "") -> str:
        project = {
            'title': project_title,
            'description': project_description,
            'dataset_hashes': dataset_hashes
        }

        return self.querier.basic_setter(Project, uid=None, payload=project)

    def create_project_api_key(self, project_hash: str, api_key_title: str, scopes: List[APIKeyScopes]) -> str:
        payload = {
            'title': api_key_title,
            'scopes': list(map(lambda scope: scope.value, scopes))
        }

        return self.querier.basic_setter(ProjectAPIKey, uid=project_hash, payload=payload)

    def get_project_api_keys(self, project_hash: str) -> List[ProjectAPIKey]:
        return self.querier.get_multiple(ProjectAPIKey, uid=project_hash)

